@@ -41,7 +41,7 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
     * Some more documentation and examples would be nice.
     """
 
-    def __init__(self, subject: [Subject,ailment.Block,Block,Function]=None, func_graph=None, max_iterations=3,
+    def __init__(self, subject: Union[Subject,ailment.Block,Block,Function]=None, func_graph=None, max_iterations=3,
                  track_tmps=False, observation_points=None, init_state: ReachingDefinitionsState=None, cc=None,
                  function_handler=None, call_stack: Optional[List[int]]=None, maximum_local_call_depth=5,
                  observe_all=False, visited_blocks=None, dep_graph: Optional['DepGraph']=None, observe_callback=None,
@@ -67,7 +67,7 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
                                                 leading to the analysed subject, from older to newer calls. Setting it
                                                 to None to limit the analysis to a single function and disable call
                                                 stack tracking; In that case, all contexts in CodeLocation will be
-                                                empty.
+                                                None, which makes CodeLocation objects contextless.
         :param int maximum_local_call_depth:    Maximum local function recursion depth.
         :param Boolean observe_all:             Observe every statement, both before and after.
         :param visited_blocks:                  A set of previously visited blocks.
@@ -100,7 +100,7 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
         else:
             self._function_handler = function_handler.hook(self)
 
-        self._call_stack: List[int] = [ ]
+        self._call_stack: Optional[List[int]] = None
         if call_stack is not None:
             self._call_stack = self._init_call_stack(call_stack, subject)
 
@@ -141,10 +141,10 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
         if self._subject.type == SubjectType.Function:
             return call_stack + [subject.addr]
         elif self._subject.type == SubjectType.Block:
-            if 'CFGFast' not in self.kb.cfgs:
+            cfg = self.kb.cfgs.get_most_accurate()
+            if cfg is None:
                 # no CFG exists
                 return call_stack
-            cfg = self.kb.cfgs['CFGFast']
             cfg_node = cfg.get_any_node(subject.addr)
             if cfg_node is None:
                 # we don't know which function this node belongs to
@@ -298,8 +298,9 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
                 canonical_size=self._canonical_size,
             )
 
-    def _merge_states(self, node, *states):
-        return states[0].merge(*states[1:])
+    def _merge_states(self, node, *states: ReachingDefinitionsState):
+        merged_state, merge_occurred = states[0].merge(*states[1:])
+        return merged_state, not merge_occurred
 
     def _run_on_node(self, node, state: ReachingDefinitionsState):
         """
@@ -346,7 +347,7 @@ class ReachingDefinitionsAnalysis(ForwardAnalysis, Analysis):  # pylint:disable=
 
         # update all definitions and all uses
         self.all_definitions |= state.all_definitions
-        for use in [state.stack_uses, state.register_uses]:
+        for use in [state.stack_uses, state.heap_uses, state.register_uses]:
             self.all_uses.merge(use)
 
         if self._node_iterations[block_key] < self._max_iterations:
