@@ -77,6 +77,7 @@ class SimFunctionArgument:
     def check_value(self, value):
         if not isinstance(value, claripy.ast.Base) and self.size is None:
             raise TypeError("Only claripy objects may be stored through SimFunctionArgument when size is not provided")
+        # TODO: this also looks byte related, change to ARCH.byte_width
         if self.size is not None and isinstance(value, claripy.ast.Base) and self.size*8 < value.length:
             raise TypeError("%s doesn't fit in an argument of size %d" % (value, self.size))
 
@@ -128,8 +129,8 @@ class SimRegArg(SimFunctionArgument):
     def set_value(self, state, value, endness=None, size=None, **kwargs):  # pylint: disable=unused-argument,arguments-differ
         self.check_value(value)
         if endness is None: endness = state.arch.register_endness
-        if isinstance(value, int): value = claripy.BVV(value, self.size*8)
-        if size is None: size = min(self.size, value.length // 8)
+        if isinstance(value, int): value = claripy.BVV(value, self.size*state.arch.byte_width)
+        if size is None: size = min(self.size, value.length // state.arch.byte_width)
         offset = self._fix_offset(state, size)
         state.registers.store(offset, value, endness=endness, size=size)
 
@@ -164,8 +165,8 @@ class SimStackArg(SimFunctionArgument):
         self.check_value(value)
         if endness is None: endness = state.arch.memory_endness
         if stack_base is None: stack_base = state.regs.sp
-        if isinstance(value, int): value = claripy.BVV(value, self.size*8)
-        state.memory.store(stack_base + self.stack_offset, value, endness=endness, size=value.length//8)
+        if isinstance(value, int): value = claripy.BVV(value, self.size*state.arch.byte_width)
+        state.memory.store(stack_base + self.stack_offset, value, endness=endness, size=value.length//state.arch.byte_width)
 
     def get_value(self, state, endness=None, stack_base=None, size=None):  # pylint: disable=arguments-differ
         if endness is None: endness = state.arch.memory_endness
@@ -175,7 +176,7 @@ class SimStackArg(SimFunctionArgument):
 
 class SimComboArg(SimFunctionArgument):
     def __init__(self, locations):
-        super(SimComboArg, self).__init__(sum(x.size for x in locations))
+        super().__init__(sum(x.size for x in locations))
         self.locations = locations
 
     def __repr__(self):
@@ -238,7 +239,7 @@ class ArgSession:
                 arg = next(self.real_args)
                 if is_fp and self.cc.is_fp_arg(arg) is False:
                     raise TypeError("Can't put a float here - concrete arg positions are specified")
-                elif not is_fp and self.cc.is_fp_arg(arg) is True:
+                if not is_fp and self.cc.is_fp_arg(arg) is True:
                     raise TypeError("Can't put an int here - concrete arg positions are specified")
             except StopIteration:
                 raise TypeError("Accessed too many arguments - concrete number are specified")
@@ -348,12 +349,12 @@ class SimCC:
     # Here are all the things a subclass needs to specify!
     #
 
-    ARG_REGS: List[str] = None                                  # A list of all the registers used for integral args, in order (names or offsets)
-    FP_ARG_REGS: List[str] = None                               # A list of all the registers used for floating point args, in order
+    ARG_REGS: List[str] = []                                    # A list of all the registers used for integral args, in order (names or offsets)
+    FP_ARG_REGS: List[str] = []                                 # A list of all the registers used for floating point args, in order
     STACKARG_SP_BUFF = 0                                        # The amount of stack space reserved between the saved return address
                                                                 # (if applicable) and the arguments. Probably zero.
     STACKARG_SP_DIFF = 0                                        # The amount of stack space reserved for the return address
-    CALLER_SAVED_REGS: List[str] = None                         # Caller-saved registers
+    CALLER_SAVED_REGS: List[str] = []                           # Caller-saved registers
     RETURN_ADDR: SimFunctionArgument = None                     # The location where the return address is stored, as a SimFunctionArgument
     RETURN_VAL: SimFunctionArgument = None                      # The location where the return value is stored, as a SimFunctionArgument
     OVERFLOW_RETURN_VAL: Optional[SimFunctionArgument] = None   # The second half of the location where a double-length return value is stored
@@ -505,8 +506,7 @@ class SimCC:
                 args = [ a.with_arch(self.arch) for a in self.func_ty.args ]
             else:
                 args = self.args
-            is_fp = [ True if isinstance(arg, (SimTypeFloat, SimTypeDouble)) or self.is_fp_arg(arg) else False
-                      for arg in args ]
+            is_fp = [ isinstance(arg, (SimTypeFloat, SimTypeDouble)) or self.is_fp_arg(arg) for arg in args ]
             if sizes is None:
                 # initialize sizes from args
                 sizes = [ ]
@@ -515,7 +515,7 @@ class SimCC:
                         if isinstance(a, SimTypeFixedSizeArray) or a.size is NotImplemented:
                             sizes.append(self.arch.bytes)
                         else:
-                            sizes.append(a.size // 8)  # SimType.size is in bits
+                            sizes.append(a.size // self.arch.byte_width)  # SimType.size is in bits
                     elif isinstance(a, SimFunctionArgument):
                         sizes.append(a.size)  # SimFunctionArgument.size is in bytes
                     else:
@@ -568,8 +568,7 @@ class SimCC:
             if self.args is None:
                 if self.func_ty is None:
                     raise ValueError("You must either customize this CC or pass a value to is_fp!")
-                else:
-                    arg_locs = self.arg_locs([False]*len(self.func_ty.args))
+                arg_locs = self.arg_locs([False]*len(self.func_ty.args))
             else:
                 arg_locs = self.args
 
@@ -686,7 +685,7 @@ class SimCC:
         allocator.apply(state, alloc_base)
 
         for loc, val in zip(arg_locs, vals):
-            if val.length > loc.size * 8:
+            if val.length > loc.size * state.arch.byte_width:
                 raise ValueError("Can't fit value {} into location {}".format(repr(val), repr(loc)))
             loc.set_value(state, val, endness='Iend_BE', stack_base=stack_base)
         self.return_addr.set_value(state, ret_addr, stack_base=stack_base)
@@ -767,8 +766,8 @@ class SimCC:
         ty = self.func_ty.returnty if self.func_ty is not None else None
         try:
             betterval = self._standardize_value(val, ty, state, None)
-        except AttributeError:
-            raise ValueError("Can't fit value %s into a return value" % repr(val))
+        except AttributeError as ex:
+            raise ValueError("Can't fit value %s into a return value" % repr(val)) from ex
 
         if self.ret_val is not None:
             loc = self.ret_val
@@ -1025,7 +1024,7 @@ class SimLyingRegArg(SimRegArg):
     """
     def __init__(self, name):
         # TODO: This looks byte-related.  Make sure to use Arch.byte_width
-        super(SimLyingRegArg, self).__init__(name, 8)
+        super().__init__(name, 8)
 
     def get_value(self, state, size=None, endness=None, **kwargs):  # pylint:disable=arguments-differ
         #val = super(SimLyingRegArg, self).get_value(state, **kwargs)
@@ -1062,6 +1061,13 @@ class SimCCCdecl(SimCC):
 class SimCCStdcall(SimCCCdecl):
     CALLEE_CLEANUP = True
 
+class SimCCMicrosoftFastcall(SimCC):
+    ARG_REGS = ['ecx', 'edx'] # Remaining arguments are passed in stack
+    STACKARG_SP_DIFF = 4 # Return address is pushed on to stack by call
+    RETURN_VAL = SimRegArg('eax', 4)
+    RETURN_ADDR = SimStackArg(0, 4)
+    ARCH = archinfo.ArchX86
+
 class SimCCMicrosoftAMD64(SimCC):
     ARG_REGS = ['rcx', 'rdx', 'r8', 'r9']
     FP_ARG_REGS = ['xmm0', 'xmm1', 'xmm2', 'xmm3']
@@ -1080,10 +1086,46 @@ class SimCCSyscall(SimCC):
     The base class of all syscall CCs.
     """
 
+    ERROR_REG : SimRegArg = None
+    SYSCALL_ERRNO_START = None
+
     @staticmethod
     def syscall_num(state) -> int:
         raise NotImplementedError()
 
+    def linux_syscall_update_error_reg(self, state, expr):
+        # special handling for Linux syscalls: on some architectures (mips/a3, powerpc/cr0_0) a bool indicating success
+        # or failure of a system call is used as an error flag (0 for success, 1 for error). we have to set this
+        if state.project is None or state.project.simos is None or state.project.simos.name != 'Linux':
+            return expr
+        if type(expr) is int:
+            expr = claripy.BVV(expr, state.arch.bits)
+        try:
+            expr = expr.ast
+        except AttributeError:
+            pass
+        nbits = self.ERROR_REG.size * state.arch.byte_width
+        error_cond = claripy.UGE(expr, self.SYSCALL_ERRNO_START)
+        if state.solver.is_false(error_cond):
+            # guaranteed no error
+            error_reg_val = claripy.BVV(0, nbits)
+        elif state.solver.is_true(error_cond):
+            # guaranteed error
+            error_reg_val = claripy.BVV(-1, nbits)
+            expr = -expr
+        else:
+            # both are satisfied, handle gracefully
+            error_reg_val = claripy.If(error_cond, claripy.BVV(-1, nbits), 0)
+            expr = claripy.If(error_cond, -expr, expr)
+
+        self.ERROR_REG.set_value(state, error_reg_val)
+        return expr
+
+    def set_return_val(self, state, val, is_fp=None, size=None, stack_base=None):
+        if self.ERROR_REG is not None:
+            val = self.linux_syscall_update_error_reg(state, val)
+        super().set_return_val(state, val, is_fp=is_fp, size=size, stack_base=stack_base)
+
 
 class SimCCX86LinuxSyscall(SimCCSyscall):
     ARG_REGS = ['ebx', 'ecx', 'edx', 'esi', 'edi', 'ebp']
@@ -1133,7 +1175,7 @@ class SimCCSystemVAMD64(SimCC):
     STACK_ALIGNMENT = 16
 
     def __init__(self, arch, args=None, ret_val=None, sp_delta=None, func_ty=None):
-        super(SimCCSystemVAMD64, self).__init__(arch, args, ret_val, sp_delta, func_ty)
+        super().__init__(arch, args, ret_val, sp_delta, func_ty)
 
         # Remove the ret address on stack
         if self.args is not None:
@@ -1266,6 +1308,9 @@ class SimCCO32LinuxSyscall(SimCCSyscall):
     RETURN_ADDR = SimRegArg('ip_at_syscall', 4)
     ARCH = archinfo.ArchMIPS32
 
+    ERROR_REG = SimRegArg('a3', 4)
+    SYSCALL_ERRNO_START = -1133
+
     @classmethod
     def _match(cls, arch, args, sp_delta):  # pylint: disable=unused-argument
         # never appears anywhere except syscalls
@@ -1275,29 +1320,6 @@ class SimCCO32LinuxSyscall(SimCCSyscall):
     def syscall_num(state):
         return state.regs.v0
 
-    @staticmethod
-    def linux_syscall_update_a3(state, expr):
-        # special handling for Linux MIPS syscalls: $a3 is used as an error flag (0 for success, 1 for error)
-        if state.project is not None and state.project.simos is not None and state.project.simos.name == 'Linux':
-            if state.solver.symbolic(expr):
-                state.regs.a3 = 0
-            else:
-                expr_val = state.solver.eval(expr)
-                if state.arch.bits == 32 and expr_val >= -1133 & 0xffff_ffff:
-                    expr_val = -(-0x1_0000_0000 + expr_val)
-                    state.regs.a3 = 1
-                elif state.arch.bits == 64 and expr_val >= -1133 & 0xffff_ffff_ffff_ffff:
-                    expr_val = -(-0x1_0000_0000_0000_0000 + expr_val)
-                    state.regs.a3 = 1
-                else:
-                    state.regs.a3 = 0
-                expr = state.solver.BVV(expr_val, state.arch.bits)
-        return expr
-
-    def set_return_val(self, state, val, is_fp=None, size=None, stack_base=None):
-        new_val = SimCCO32LinuxSyscall.linux_syscall_update_a3(state, val)
-        super().set_return_val(state, new_val, is_fp=is_fp, size=size, stack_base=stack_base)
-
 
 class SimCCO64(SimCC):      # TODO: add n32 and n64
     ARG_REGS = [ 'a0', 'a1', 'a2', 'a3' ]
@@ -1315,6 +1337,9 @@ class SimCCO64LinuxSyscall(SimCCSyscall):
     RETURN_ADDR = SimRegArg('ip_at_syscall', 8)
     ARCH = archinfo.ArchMIPS64
 
+    ERROR_REG = SimRegArg('a3', 8)
+    SYSCALL_ERRNO_START = -1133
+
     @classmethod
     def _match(cls, arch, args, sp_delta):  # pylint: disable=unused-argument
         # never appears anywhere except syscalls
@@ -1324,10 +1349,6 @@ class SimCCO64LinuxSyscall(SimCCSyscall):
     def syscall_num(state):
         return state.regs.v0
 
-    def set_return_val(self, state, val, is_fp=None, size=None, stack_base=None):
-        new_val = SimCCO32LinuxSyscall.linux_syscall_update_a3(state, val)
-        super().set_return_val(state, new_val, is_fp=is_fp, size=size, stack_base=stack_base)
-
 
 class SimCCPowerPC(SimCC):
     ARG_REGS = [ 'r3', 'r4', 'r5', 'r6', 'r7', 'r8', 'r9', 'r10' ]
@@ -1346,6 +1367,9 @@ class SimCCPowerPCLinuxSyscall(SimCCSyscall):
     RETURN_ADDR = SimRegArg('ip_at_syscall', 4)
     ARCH = archinfo.ArchPPC32
 
+    ERROR_REG = SimRegArg('cr0_0', 1)
+    SYSCALL_ERRNO_START = -515
+
     @classmethod
     def _match(cls, arch, args, sp_delta):  # pylint: disable=unused-argument
         # never appears anywhere except syscalls
@@ -1355,29 +1379,6 @@ class SimCCPowerPCLinuxSyscall(SimCCSyscall):
     def syscall_num(state):
         return state.regs.r0
 
-    @staticmethod
-    def linux_syscall_update_cr0(state, expr):
-        # special handling for Linux MIPS syscalls: $cr0_0 is used as an error flag (0 for success, 1 for error)
-        if state.project is not None and state.project.simos is not None and state.project.simos.name == 'Linux':
-            if state.solver.symbolic(expr):
-                state.regs.cr0_0 = 0
-            else:
-                expr_val = state.solver.eval(expr)
-                if state.arch.bits == 32 and expr_val > -515 & 0xffff_ffff:
-                    expr_val = -(-0x1_0000_0000 + expr_val)
-                    state.regs.cr0_0 = 1
-                elif state.arch.bits == 64 and expr_val > -515 & 0xffff_ffff_ffff_ffff:
-                    expr_val = -(-0x1_0000_0000_0000_0000 + expr_val)
-                    state.regs.cr0_0 = 1
-                else:
-                    state.regs.cr0_0 = 0
-                expr = state.solver.BVV(expr_val, state.arch.bits)
-        return expr
-
-    def set_return_val(self, state, val, is_fp=None, size=None, stack_base=None):
-        new_val = SimCCPowerPCLinuxSyscall.linux_syscall_update_cr0(state, val)
-        super().set_return_val(state, new_val, is_fp=is_fp, size=size, stack_base=stack_base)
-
 
 class SimCCPowerPC64(SimCC):
     ARG_REGS = [ 'r3', 'r4', 'r5', 'r6', 'r7', 'r8', 'r9', 'r10' ]
@@ -1396,6 +1397,9 @@ class SimCCPowerPC64LinuxSyscall(SimCCSyscall):
     RETURN_ADDR = SimRegArg('ip_at_syscall', 8)
     ARCH = archinfo.ArchPPC64
 
+    ERROR_REG = SimRegArg('cr0_0', 1)
+    SYSCALL_ERRNO_START = -515
+
     @classmethod
     def _match(cls, arch, args, sp_delta):  # pylint: disable=unused-argument
         # never appears anywhere except syscalls
@@ -1405,10 +1409,6 @@ class SimCCPowerPC64LinuxSyscall(SimCCSyscall):
     def syscall_num(state):
         return state.regs.r0
 
-    def set_return_val(self, state, val, is_fp=None, size=None, stack_base=None):
-        new_val = SimCCPowerPCLinuxSyscall.linux_syscall_update_cr0(state, val)
-        super().set_return_val(state, new_val, is_fp=is_fp, size=size, stack_base=stack_base)
-
 
 class SimCCSoot(SimCC):
     ARCH = archinfo.ArchSoot
