@@ -134,7 +134,7 @@ arithmetic_operation_map = {
     'Add': '__add__',
     'Sub': '__sub__',
     'Mul': '__mul__',
-    'Div': '__div__',
+    'Div': '__floordiv__',
     'Neg': 'Neg',
     'Abs': 'Abs',
     'Mod': '__mod__',
@@ -430,7 +430,7 @@ class SimIROp:
         else:
             raise SimOperationError("op_mapped called with invalid mapping, for %s" % self.name)
 
-        if o == '__div__' and self.is_signed:
+        if o == '__floordiv__' and self.is_signed:
             # yikes!!!!!!!
             return claripy.SDiv(*sized_args)
 
@@ -1053,7 +1053,7 @@ class SimIROp:
         ordered_0 = list(reversed(args[0].chop(self._vector_size)))
         ordered_1 = list(reversed(args[1].chop(self._vector_size)))
         res = []
-        nbits = int(math.log2(self._vector_size))
+        nbits = int(math.log2(self._vector_count))
         for pword in ordered_1:
             switch = pword[nbits-1:0]
             kill = pword[self._vector_size - 1]
