@@ -11,3 +11,5 @@ from .plugin import KnowledgeBasePlugin
 from .sync import SynchronizationManager
 from .patches import PatchManager
 from .key_definitions import KeyDefinitionManager
+from .propagations import Propagations
+from .structured_code import StructuredCodeManager
