@@ -1,7 +1,9 @@
-from typing import Set, List, Tuple, Dict, Union, TYPE_CHECKING
+from typing import Set, List, Tuple, Dict, Union, Optional, TYPE_CHECKING
 import logging
 from collections import defaultdict
-from itertools import count
+from itertools import count, chain
+
+import networkx
 
 from claripy.utils.orderedset import OrderedSet
 
@@ -63,6 +65,9 @@ class VariableManagerInternal:
             'global': count(),
         }
 
+        self._unified_variables: Set[SimVariable] = set()
+        self._variables_to_unified_variables: Dict[SimVariable, SimVariable] = { }
+
         self._phi_variables = { }
         self._phi_variables_by_block = defaultdict(set)
 
@@ -100,15 +105,25 @@ class VariableManagerInternal:
         else:
             raise ValueError('Unsupported sort %s in add_variable().' % sort)
 
-    def set_variable(self, sort, start, variable):
+    def set_variable(self, sort, start, variable: SimVariable):
         if sort == 'stack':
-            self._stack_region.set_variable(start, variable)
+            region = self._stack_region
         elif sort == 'register':
-            self._register_region.set_variable(start, variable)
+            region = self._register_region
         elif sort == 'global':
-            self._global_region.set_variable(start, variable)
+            region = self._global_region
         else:
-            raise ValueError('Unsupported sort %s in add_variable().' % sort)
+            raise ValueError('Unsupported sort %s in set_variable().' % sort)
+        existing = [x for x in region.get_variables_by_offset(start) if x.ident == variable.ident]
+        if len(existing) == 1:
+            var = existing[0]
+            if var.name is not None and not variable.renamed:
+                variable.name = var.name
+                variable.renamed = var.renamed
+        else:
+            # implicitly overwrite or add I guess
+            pass
+        region.set_variable(start, variable)
 
     def write_to(self, variable, offset, location, overwrite=False, atom=None):
         self._record_variable_access('write', variable, offset, location, overwrite=overwrite, atom=atom)
@@ -168,7 +183,7 @@ class VariableManagerInternal:
             ident_sort = 'register'
             a = SimRegisterVariable(repre.reg, repre.size, ident=self.next_variable_ident(ident_sort))
         elif repre_type is SimMemoryVariable:
-            ident_sort = 'memory'
+            ident_sort = 'global'
             a = SimMemoryVariable(repre.addr, repre.size, ident=self.next_variable_ident(ident_sort))
         elif repre_type is SimStackVariable:
             ident_sort = 'stack'
@@ -365,9 +380,10 @@ class VariableManagerInternal:
 
     def assign_variable_names(self, labels=None):
         """
-        Assign default names to all variables.
+        Assign default names to all SSA variables.
 
-        :return: None
+        :param labels:  Known labels in the binary.
+        :return:        None
         """
 
         for var in self._variables:
@@ -388,8 +404,81 @@ class VariableManagerInternal:
                     continue
                 if labels is not None and var.addr in labels:
                     var.name = labels[var.addr]
-                else:
+                    if "@@" in var.name:
+                        var.name = var.name[:var.name.index("@@")]
+                elif var.ident is not None:
                     var.name = var.ident
+                else:
+                    var.name = "g_%x" % var.addr
+
+    def assign_unified_variable_names(self, labels=None, reset:bool=False):
+        """
+        Assign default names to all unified variables.
+
+        :param labels:  Known labels in the binary.
+        :param reset:   Reset all variable names or not.
+        :return:        None
+        """
+
+        if not self._unified_variables:
+            return
+
+        sorted_stack_variables = [ ]
+        sorted_reg_variables = [ ]
+        arg_vars = [ ]
+
+        for var in self._unified_variables:
+            if isinstance(var, SimStackVariable):
+                if var.ident and var.ident.startswith('iarg_'):
+                    arg_vars.append(var)
+                else:
+                    sorted_stack_variables.append(var)
+
+            elif isinstance(var, SimRegisterVariable):
+                if var.ident and var.ident.startswith('arg_'):
+                    arg_vars.append(var)
+                else:
+                    sorted_reg_variables.append(var)
+
+            elif isinstance(var, SimMemoryVariable):
+                if not reset and var.name is not None:
+                    continue
+                # assign names directly
+                if labels is not None and var.addr in labels:
+                    var.name = labels[var.addr]
+                    if "@@" in var.name:
+                        var.name = var.name[:var.name.index("@@")]
+                elif var.ident:
+                    var.name = var.ident
+                else:
+                    var.name = f"g_{var.addr:x}"
+
+        # rename variables in a fixed order
+        var_ctr = count(0)
+
+        sorted_stack_variables = sorted(sorted_stack_variables, key=lambda v: v.offset)
+        sorted_reg_variables = sorted(sorted_reg_variables, key=lambda v: v.reg)
+
+        for var in chain(sorted_stack_variables, sorted_reg_variables):
+            idx = next(var_ctr)
+            if var.name is not None and not reset:
+                continue
+            if isinstance(var, SimStackVariable):
+                var.name = f'v{idx}'
+            elif isinstance(var, SimRegisterVariable):
+                var.name = f"v{idx}"
+            # clear the hash cache
+            var._hash = None
+
+        # rename arguments but keeping the original order
+        arg_ctr = count(0)
+        arg_vars = sorted(arg_vars, key=lambda v: int(v.ident[v.ident.index("_")+1:]) if v.ident else 0)
+        for var in arg_vars:
+            idx = next(arg_ctr)
+            if var.name is not None and not reset:
+                continue
+            var.name = f"a{idx}"
+            var._hash = None
 
     def get_variable_type(self, var):
         return self.types.get(var, None)
@@ -397,13 +486,85 @@ class VariableManagerInternal:
     def remove_types(self):
         self.types.clear()
 
+    def unify_variables(self) -> None:
+        """
+        Map SSA variables to a unified variable. Fill in self._unified_variables.
+        """
+
+        stack_vars: Dict[int,List[SimStackVariable]] = defaultdict(list)
+        reg_vars: Set[SimRegisterVariable] = set()
+
+        # unify stack variables based on their locations
+        for v in self.get_variables():
+            if isinstance(v, SimStackVariable):
+                stack_vars[v.offset].append(v)
+            elif isinstance(v, SimRegisterVariable):
+                reg_vars.add(v)
+
+        for _, vs in stack_vars.items():
+            unified = vs[0].copy()
+            for v in vs:
+                self.set_unified_variable(v, unified)
+
+        # unify register variables based on phi nodes
+        graph = networkx.Graph()
+        for v, subvs in self._phi_variables.items():
+            if not isinstance(v, SimRegisterVariable):
+                continue
+            if not self.get_variable_accesses(v):
+                # this phi node has never been used - discard it
+                continue
+            for subv in subvs:
+                graph.add_edge(v, subv)
+
+        for nodes in networkx.connected_components(graph):
+            if len(nodes) <= 1:
+                continue
+            nodes = list(nodes)
+            unified = nodes[0].copy()
+            for v in nodes:
+                self.set_unified_variable(v, unified)
+            for v in nodes:
+                reg_vars.discard(v)
+
+        for v in reg_vars:
+            self.set_unified_variable(v, v)
+
+    def set_unified_variable(self, variable: SimVariable, unified: SimVariable) -> None:
+        """
+        Set the unified variable for a given SSA variable.
+
+        :param variable:    The SSA variable.
+        :param unified:     The unified variable.
+        :return:            None
+        """
+        old_unified = self._variables_to_unified_variables.get(variable, None)
+        if old_unified is not None and old_unified is not unified:
+            self._unified_variables.discard(old_unified)
+            if old_unified.name is not None and not unified.renamed:
+                unified.name = old_unified.name
+                unified.renamed = old_unified.renamed
+
+        self._unified_variables.add(unified)
+        self._variables_to_unified_variables[variable] = unified
+
+    def unified_variable(self, variable: SimVariable) -> Optional[SimVariable]:
+        """
+        Return the unified variable for a given SSA variable,
+
+        :param variable:    The SSA variable.
+        :return:            The unified variable, or None if there is no such SSA variable.
+        """
+
+        return self._variables_to_unified_variables.get(variable, None)
+
 
 class VariableManager(KnowledgeBasePlugin):
     """
     Manage variables.
     """
     def __init__(self, kb):
-        super(VariableManager, self).__init__()
+        super().__init__()
         self._kb: 'KnowledgeBase' = kb
         self.global_manager = VariableManagerInternal(self)
         self.function_managers: Dict[int,VariableManagerInternal] = { }
@@ -442,7 +603,9 @@ class VariableManager(KnowledgeBasePlugin):
         return key in self.function_managers
 
     def get_function_manager(self, func_addr) -> VariableManagerInternal:
-        if not isinstance(func_addr, int):
+        if isinstance(func_addr, str):
+            func_addr = self._kb.labels.lookup(func_addr)
+        elif not isinstance(func_addr, int):
             raise TypeError('Argument "func_addr" must be an int.')
 
         if func_addr not in self.function_managers:
