@@ -1,13 +1,15 @@
 import inspect
 import copy
 import itertools
-from cle import Symbol
-
 import logging
-l = logging.getLogger(name=__name__)
-
+from typing import TYPE_CHECKING
+from cle import SymbolType
 from archinfo.arch_soot import SootAddressDescriptor
 
+if TYPE_CHECKING:
+    import angr
+
+l = logging.getLogger(name=__name__)
 symbolic_count = itertools.count()
 
 
@@ -18,7 +20,7 @@ class SimProcedure:
     You may subclass SimProcedure and override ``run()``, replacing it with mutating ``self.state`` however you like,
     and then either returning a value or jumping away somehow.
 
-    A detailed discussion of programming SimProcedures may be found at https://docs.angr.io/docs/simprocedures.md
+    A detailed discussion of programming SimProcedures may be found at https://docs.angr.io/extending-angr/simprocedures
 
     :param arch:            The architecture to use for this procedure
 
@@ -30,9 +32,61 @@ class SimProcedure:
     :param is_syscall:      Whether this procedure is a syscall
     :param num_args:        The number of arguments this procedure should extract
     :param display_name:    The name to use when displaying this procedure
+    :param library_name:    The name of the library from which the function we're emulating comes
     :param cc:              The SimCC to use for this procedure
     :param sim_kwargs:      Additional keyword arguments to be passed to run()
     :param is_function:     Whether this procedure emulates a function
+
+    The following class variables should be set if necessary when implementing a new SimProcedure:
+
+    :cvar NO_RET:           Set this to true if control flow will never return from this function
+    :cvar ADDS_EXITS:       Set this to true if you do any control flow other than returning
+    :cvar IS_FUNCTION:      Does this procedure simulate a function? True by default
+    :cvar ARGS_MISMATCH:    Does this procedure have a different list of arguments than what is provided in the
+                            function specification? This may happen when we manually extract arguments in the run()
+                            method of a SimProcedure. False by default.
+    :cvar local_vars:       If you use ``self.call()``, set this to a list of all the local variable
+                            names in your class. They will be restored on return.
+
+    The following instance variables are available when working with simprocedures from the inside or the outside:
+
+    :ivar project:          The associated angr project
+    :ivar arch:             The associated architecture
+    :ivar addr:             The linear address at which the procedure is executing
+    :ivar cc:               The calling convention in use for engaging with the ABI
+    :ivar canonical:        The canonical version of this SimProcedure. Procedures are deepcopied for many reasons,
+                            including to be able to store state related to a specific run and to be able to hook
+                            continuations.
+    :ivar kwargs:           Any extra keyword arguments used to construct the procedure; will be passed to ``run``
+    :ivar display_name:     See the eponymous parameter
+    :ivar library_name:     See the eponymous parameter
+    :ivar abi:
+    :ivar symbolic_return:  See the eponymous parameter
+    :ivar syscall_number:   If this procedure is a syscall, the number will be populated here.
+    :ivar returns:          See eponymous parameter and NO_RET cvar
+    :ivar is_syscall:       See eponymous parameter
+    :ivar is_function:      See eponymous parameter and cvar
+    :ivar is_stub:          See eponymous parameter
+    :ivar is_continuation:  Whether this procedure is the original or a continuation resulting from ``self.call()``
+    :ivar continuations:    A mapping from name to each known continuation
+    :ivar run_func:         The name of the function implementing the procedure. "run" by default, but different in
+                            continuations.
+    :ivar num_args:         The number of arguments to the procedure. If not provided in the parameter, extracted from
+                            the definition of ``self.run``
+
+    The following instance variables are only used in a copy of the procedure that is actually executing on a state:
+
+    :ivar state:            The SimState we should be mutating to perform the procedure
+    :ivar successors:       The SimSuccessors associated with the current step
+    :ivar arguments:        The function arguments, deserialized from the state
+    :ivar use_state_arguments:
+                            Whether we're using arguments extracted from the state or manually provided
+    :ivar ret_to:           The current return address
+    :ivar ret_expr:         The computed return value
+    :ivar call_ret_expr:    The return value from having used ``self.call()``
+    :ivar inhibit_autoret:  Whether we should avoid automatically adding an exit for returning once the run function
+                            ends
+
     """
     def __init__(
         self, project=None, cc=None, symbolic_return=None,
@@ -41,10 +95,10 @@ class SimProcedure:
         is_function=None, **kwargs
     ):
         # WE'LL FIGURE IT OUT
-        self.project = project
+        self.project = project # type: angr.Project
         self.arch = project.arch if project is not None else None
         self.addr = None
-        self.cc = cc
+        self.cc = cc # type: angr.SimCC
         self.canonical = self
 
         self.kwargs = kwargs
@@ -54,10 +108,6 @@ class SimProcedure:
         self.abi = None
         self.symbolic_return = symbolic_return
 
-        # types
-        self.argument_types = { } # a dictionary of index-to-type (i.e., type of arg 0: SimTypeString())
-        self.return_type = None
-
         # set some properties about the type of procedure this is
         self.returns = returns if returns is not None else not self.NO_RET
         self.is_syscall = is_syscall
@@ -107,7 +157,7 @@ class SimProcedure:
         provide arguments to the function.
         """
         # fill out all the fun stuff we don't want to frontload
-        if self.addr is None and not state.regs.ip.symbolic:
+        if self.addr is None and not state.regs._ip.symbolic:
             self.addr = state.addr
         if self.arch is None:
             self.arch = state.arch
@@ -127,30 +177,26 @@ class SimProcedure:
         inst.inhibit_autoret = False
 
         # check to see if this is a syscall and if we should override its return value
-        override = None
         if inst.is_syscall:
             state.history.recent_syscall_count = 1
-            if len(state.posix.queued_syscall_returns):
-                override = state.posix.queued_syscall_returns.pop(0)
 
-        if callable(override):
-            try:
-                r = override(state, run=inst)
-            except TypeError:
-                r = override(state)
-            inst.use_state_arguments = True
+        state._inspect(
+            'simprocedure',
+            BP_BEFORE,
+            simprocedure_name=inst.display_name,
+            simprocedure_addr=self.addr,
+            simprocedure=inst,
+            simprocedure_result=NO_OVERRIDE
+        )
 
-        elif override is not None:
-            r = override
-            inst.use_state_arguments = True
-
-        else:
+        r = state._inspect_getattr('simprocedure_result', NO_OVERRIDE)
+        if r is NO_OVERRIDE:
             # get the arguments
 
             # If the simprocedure is related to a Java function call the appropriate setup_args methos
             # TODO: should we move this?
             if self.is_java:
-                sim_args = self._setup_args(inst, state, arguments)
+                sim_args = self._setup_args(inst, state, arguments) #pylint:disable=assignment-from-no-return
                 self.use_state_arguments = False
 
             # handle if this is a continuation from a return
@@ -158,7 +204,9 @@ class SimProcedure:
                 if state.callstack.top.procedure_data is None:
                     raise SimProcedureError("Tried to return to a SimProcedure in an inapplicable stack frame!")
 
-                saved_sp, sim_args, saved_local_vars, saved_lr = state.callstack.top.procedure_data
+                saved_sp, sim_args, saved_local_vars, saved_lr, ideal_addr = state.callstack.top.procedure_data
+                if ideal_addr != inst.addr:
+                    raise SimShadowStackError("I can't emulate this consequence of stack smashing")
                 state.regs.sp = saved_sp
                 if saved_lr is not None:
                     state.regs.lr = saved_lr
@@ -181,6 +229,13 @@ class SimProcedure:
             l.debug("Executing %s%s%s%s%s with %s, %s", *(inst._describe_me() + (sim_args, inst.kwargs)))
             r = getattr(inst, inst.run_func)(*sim_args, **inst.kwargs)
 
+        state._inspect(
+            'simprocedure',
+            BP_AFTER,
+            simprocedure_result=r
+        )
+        r = state._inspect_getattr('simprocedure_result', r)
+
         if inst.returns and inst.is_function and not inst.inhibit_autoret:
             inst.ret(r)
 
@@ -193,7 +248,7 @@ class SimProcedure:
             target_name = '%s.%s' % (self.display_name, name)
             should_be_none = self.project.loader.extern_object.get_symbol(target_name)
             if should_be_none is None:
-                cont.addr = self.project.loader.extern_object.make_extern(target_name, sym_type=Symbol.TYPE_OTHER).rebased_addr
+                cont.addr = self.project.loader.extern_object.make_extern(target_name, sym_type=SymbolType.TYPE_OTHER).rebased_addr
             else:
                 l.error("Trying to make continuation %s but it already exists. This is bad.", target_name)
                 cont.addr = self.project.loader.extern_object.allocate()
@@ -207,15 +262,12 @@ class SimProcedure:
     # Implement these in a subclass of SimProcedure!
     #
 
-    NO_RET = False          # set this to true if control flow will never return from this function
-    ADDS_EXITS = False      # set this to true if you do any control flow other than returning
-    IS_FUNCTION = True      # does this procedure simulate a function?
-    ARGS_MISMATCH = False   # does this procedure have a different list of arguments than what is provided in the
-                            # function specification? This may happen when we manually extract arguments in the run()
-                            # method of a SimProcedure.
-
-    local_vars = ()         # if you use self.call(), set this to a list of all the local variable
-                            # names in your class. They will be restored on return.
+    NO_RET = False
+    ADDS_EXITS = False
+    IS_FUNCTION = True
+    ARGS_MISMATCH = False
+    ALT_NAMES = None  # alternative names
+    local_vars = ()
 
     def run(self, *args, **kwargs): # pylint: disable=unused-argument
         """
@@ -229,15 +281,15 @@ class SimProcedure:
         exits for scenarios where states are not available (e.g. when building a fast CFG).
 
         :param list blocks: Blocks that are executed before reaching this SimProcedure.
-        :return: A list of tuples. Each tuple is (address, jumpkind).
+        :return: A list of dicts. Each dict should contain the following entries: 'address', 'jumpkind', and 'namehint'.
         :rtype: list
         """
 
         if self.ADDS_EXITS:
             raise SimProcedureError("static_exits() is not implemented for %s" % self)
-        else:
-            # This SimProcedure does not add any new exit
-            return [ ]
+
+        # This SimProcedure does not add any new exit
+        return []
 
     #
     # misc properties
@@ -251,6 +303,11 @@ class SimProcedure:
     # Working with calling conventions
     #
 
+    def _setup_args(self, inst, state, args): #pylint:disable=unused-argument,no-self-use
+        raise SimProcedureError("the java-specific _setup_args() method was invoked on a non-Java SimProcedure.")
+    def _compute_ret_addr(self, expr): #pylint:disable=unused-argument,no-self-use
+        raise SimProcedureError("the java-specific _compute_ret_addr() method was invoked on a non-Java SimProcedure.")
+
     def set_args(self, args):
         arg_session = self.cc.arg_session
         for arg in args:
@@ -326,7 +383,7 @@ class SimProcedure:
         # TODO: I had to put this check here because I don't understand why self.use_state_arguments gets reset to true
         # when calling the function ret. at the calling point the attribute is set to False
         if isinstance(self.addr, SootAddressDescriptor):
-            ret_addr = self._compute_ret_addr(expr)
+            ret_addr = self._compute_ret_addr(expr) #pylint:disable=assignment-from-no-return
         elif self.use_state_arguments:
             ret_addr = self.cc.teardown_callsite(
                     self.state,
@@ -371,7 +428,8 @@ class SimProcedure:
         simcallstack_entry = (self.state.regs.sp if hasattr(self.state.regs, "sp") else None,
                               self.arguments,
                               saved_local_vars,
-                              self.state.regs.lr if self.state.arch.lr_offset is not None else None)
+                              self.state.regs.lr if self.state.arch.lr_offset is not None else None,
+                              ret_addr)
         cc.setup_callsite(call_state, ret_addr, args)
         call_state.callstack.top.procedure_data = simcallstack_entry
 
@@ -435,9 +493,26 @@ class SimProcedure:
     def _prepare_ret_state(self):
         pass
 
+    @property
+    def argument_types(self):  # pylint: disable=no-self-use
+        return None
+
+    @argument_types.setter
+    def argument_types(self, v):  # pylint: disable=unused-argument,no-self-use
+        l.critical("SimProcedure.argument_types is deprecated. specify the function signature in the cc")
+
+    @property
+    def return_type(self):  # pylint: disable=no-self-use
+        return None
+
+    @return_type.setter
+    def return_type(self, v):  # pylint: disable=unused-argument,no-self-use
+        l.critical("SimProcedure.return_type is deprecated. specify the function signature in the cc")
+
 
 from . import sim_options as o
-from angr.errors import SimProcedureError, SimProcedureArgumentError
+from angr.errors import SimProcedureError, SimProcedureArgumentError, SimShadowStackError
 from angr.sim_type import SimTypePointer
 from angr.state_plugins.sim_action import SimActionExit
 from angr.calling_conventions import DEFAULT_CC
+from .state_plugins import BP_AFTER, BP_BEFORE, NO_OVERRIDE
