@@ -1,5 +1,7 @@
 import logging
 import struct
+from typing import Optional
+
 import angr # for types
 
 import claripy
@@ -16,7 +18,6 @@ from ..calling_conventions import DEFAULT_CC
 from ..procedures import SIM_PROCEDURES as P
 from .. import sim_options as o
 from ..storage.file import SimFileStream, SimFileBase
-from ..misc import IRange
 
 
 _l = logging.getLogger(name=__name__)
@@ -82,10 +83,11 @@ class SimOS:
             sym = scope.get_symbol(name)
 
         if sym is not None:
-            if self.project.is_hooked(sym.rebased_addr):
-                if not self.project.hooked_by(sym.rebased_addr).is_stub:
+            addr, _ = self.prepare_function_symbol(name, basic_addr=sym.rebased_addr)
+            if self.project.is_hooked(addr):
+                if not self.project.hooked_by(addr).is_stub:
                     return
-            self.project.hook(sym.rebased_addr, hook)
+            self.project.hook(addr, hook)
 
     def state_blank(self, addr=None, initial_prefix=None, brk=None, stack_end=None, stack_size=1024*1024*8, stdin=None,
                     thread_idx=None, permissions_backer=None, **kwargs):
@@ -157,9 +159,6 @@ class SimOS:
         actual_brk = (last_addr - last_addr % 0x1000 + 0x1000) if brk is None else brk
         state.register_plugin('posix', SimSystemPosix(stdin=stdin, brk=actual_brk))
 
-        if state.arch.sp_offset is not None:
-            state.regs.sp = stack_end
-
         if initial_prefix is not None:
             for reg in state.arch.default_symbolic_registers:
                 state.registers.store(reg, state.solver.BVS(
@@ -169,6 +168,9 @@ class SimOS:
                     key=('reg', reg),
                     eternal=True))
 
+        if state.arch.sp_offset is not None:
+            state.regs.sp = stack_end
+
         for reg, val, is_addr, mem_region in state.arch.default_register_values:
             region_base = None  # so pycharm does not complain
 
@@ -320,9 +322,12 @@ class SimOS:
     def syscall(self, state, allow_unsupported=True):
         return None
 
-    def syscall_abi(self, state):
+    def syscall_abi(self, state) -> str:
         return None
 
+    def syscall_cc(self, state) -> Optional[angr.calling_conventions.SimCCSyscall]:
+        raise NotImplementedError()
+
     def is_syscall_addr(self, addr):
         return False
 
