@@ -1,12 +1,13 @@
 import cffi
-from typing import Tuple, Type, Dict, Optional, Iterable, List, Any, Set
+from typing import Tuple, Type, Dict, Optional, Iterable, Set, Any
 import logging
 from collections import defaultdict
 
 import claripy
 
 from angr.storage.memory_mixins import MemoryMixin
-from angr.storage.memory_mixins.paged_memory.pages import PageType, ListPage, UltraPage
+from angr.storage.memory_mixins.paged_memory.pages import PageType, ListPage, UltraPage, MVListPage
+from angr.storage.memory_object import SimLabeledMemoryObject
 from ....errors import SimMemoryError
 
 # yeet
@@ -32,21 +33,16 @@ class PagedMemoryMixin(MemoryMixin):
         self._default_permissions = default_permissions
         self._pages: Dict[int, Optional[PageType]] = {}
 
-    def __del__(self):
-        # a thought: we could support mapping pages in multiple places in memory if here we just kept a set of released
-        # page ids and never released any page more than once
-        for page in self._pages.values():
-            if page is not None:
-                page.release_shared()
-
     @MemoryMixin.memo
     def copy(self, memo):
         o = super().copy(memo)
 
         o.page_size = self.page_size
-        o._pages = dict(self._pages)
-        o._permissions_map = self._permissions_map
+        o._extra_page_kwargs = self._extra_page_kwargs
+
         o._default_permissions = self._default_permissions
+        o._permissions_map = self._permissions_map
+        o._pages = dict(self._pages)
 
         for page in o._pages.values():
             if page is not None:
@@ -54,6 +50,13 @@ class PagedMemoryMixin(MemoryMixin):
 
         return o
 
+    def __del__(self):
+        # a thought: we could support mapping pages in multiple places in memory if here we just kept a set of released
+        # page ids and never released any page more than once
+        for page in self._pages.values():
+            if page is not None:
+                page.release_shared()
+
     def _get_page(self, pageno: int, writing: bool, **kwargs) -> PageType:
         force_default = True
         # force_default means don't consult any "backers"
@@ -83,7 +86,7 @@ class PagedMemoryMixin(MemoryMixin):
         if not allow_default:
             raise SimMemoryError("I have been instructed not to create a default page")
 
-        return self.PAGE_TYPE(**self._page_kwargs(pageno, permissions))
+        return self.PAGE_TYPE(**self._page_kwargs(pageno, permissions))  # pylint:disable=not-callable
 
     def _page_kwargs(self, pageno, permissions):
         # permissions lookup: let permissions arg override everything else
@@ -160,10 +163,14 @@ class PagedMemoryMixin(MemoryMixin):
 
         # fasttrack basic case
         if pageoff + size <= self.page_size:
-            sub_data = sub_gen.send(size)
-            page = self._get_page(pageno, True, **kwargs)
-            page.store(pageoff, sub_data, size=size, endness=endness, page_addr=pageno*self.page_size, memory=self,
-                       cooperate=True, **kwargs)
+            written_size = 0
+            while written_size < size:
+                sub_data, sub_data_size = sub_gen.send(size - written_size)
+                page = self._get_page(pageno, True, **kwargs)
+                sub_data_size = min(sub_data_size, size - written_size)
+                page.store(pageoff + written_size, sub_data, size=sub_data_size, endness=endness,
+                           page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs)
+                written_size += sub_data_size
             sub_gen.close()
             return
 
@@ -174,11 +181,14 @@ class PagedMemoryMixin(MemoryMixin):
             # instead of overwriting the old one. would have to be careful about maintaining the contract for _get_page
             page = self._get_page(pageno, True, **kwargs)
             sub_size = min(self.page_size-pageoff, size-bytes_done)
+            written_size = 0
 
-            sub_data = sub_gen.send(sub_size)
-
-            page.store(pageoff, sub_data, size=sub_size, endness=endness, page_addr=pageno*self.page_size, memory=self,
-                       cooperate=True, **kwargs)
+            while written_size < sub_size:
+                sub_data, sub_data_size = sub_gen.send(sub_size)
+                sub_data_size = min(sub_data_size, sub_size - written_size)
+                page.store(pageoff + written_size, sub_data, size=sub_data_size, endness=endness,
+                           page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs)
+                written_size += sub_data_size
 
             bytes_done += sub_size
             pageno = (pageno + 1) % max_pageno
@@ -187,15 +197,32 @@ class PagedMemoryMixin(MemoryMixin):
         sub_gen.close()
 
     def merge(self, others: Iterable['PagedMemoryMixin'], merge_conditions, common_ancestor=None) -> bool:
-        changed_pages = set()
+        changed_pages_and_offsets: Dict[int,Optional[Set[int]]] = {}
         for o in others:
-            changed_pages |= self.changed_pages(o)
+            for changed_page, changed_offsets in self.changed_pages(o).items():
+                if changed_offsets is None:
+                    changed_pages_and_offsets[changed_page] = None
+                else:
+                    # changed_offsets is a set of offsets (ints)
+                    if changed_page not in changed_pages_and_offsets:
+                        # update our dict
+                        changed_pages_and_offsets[changed_page] = changed_offsets
+                    else:
+                        # changed_page in our dict
+                        if changed_pages_and_offsets[changed_page] is None:
+                            # in at least one `other` memory can we not determine the changed offsets
+                            # do nothing
+                            pass
+                        else:
+                            # union changed_offsets with known ones
+                            changed_pages_and_offsets[changed_page] = \
+                                changed_pages_and_offsets[changed_page].union(changed_offsets)
 
         if merge_conditions is None:
             merge_conditions = [None] * (len(list(others)) + 1)
 
         merged_bytes = set()
-        for page_no in sorted(changed_pages):
+        for page_no in sorted(changed_pages_and_offsets.keys()):
             l.debug("... on page %x", page_no)
 
             page = self._get_page(page_no, True)
@@ -206,11 +233,13 @@ class PagedMemoryMixin(MemoryMixin):
                     other_pages.append(o._get_page(page_no, False))
 
             page_addr = page_no * self.page_size
-            merged_offsets = page.merge(other_pages, merge_conditions, page_addr=page_addr, memory=self)
+            changed_offsets = changed_pages_and_offsets[page_no]
+            merged_offsets = page.merge(other_pages, merge_conditions, page_addr=page_addr, memory=self,
+                                        changed_offsets=changed_offsets)
             for off in merged_offsets:
                 merged_bytes.add(page_addr + off)
 
-        return True if merged_bytes else False
+        return bool(merged_bytes)
 
     def permissions(self, addr, permissions=None, **kwargs):
         if type(addr) is not int:
@@ -265,7 +294,7 @@ class PagedMemoryMixin(MemoryMixin):
         except SimMemoryError:
             pass  # good, expected error
         else:
-            raise SimMemoryError("Page is already mapped")
+            raise SimMemoryError("Page is already mapped", pageno * self.page_size)
 
         page = self._initialize_default_page(pageno, permissions=permissions, **kwargs)
         self._pages[pageno] = page
@@ -273,7 +302,7 @@ class PagedMemoryMixin(MemoryMixin):
             page.store(0, None, size=self.page_size, endness='Iend_BE', page_addr=pageno*self.page_size, memory=self,
                        **kwargs)
 
-    def _unmap_page(self, pageno, **kwargs):
+    def _unmap_page(self, pageno, **kwargs):  # pylint:disable=unused-argument
         try:
             if self._pages[pageno] is not None:
                 self._pages[pageno].release_shared()
@@ -469,25 +498,26 @@ class PagedMemoryMixin(MemoryMixin):
 
         return changes
 
-    def changed_pages(self, other) -> Set[int]:
+    def changed_pages(self, other) -> Dict[int,Optional[Set[int]]]:
         my_pages = set(self._pages)
         other_pages = set(other._pages)
         intersection = my_pages.intersection(other_pages)
         difference = my_pages.difference(other_pages)
 
-        changes: Set[int] = set(difference)
+        changes: Dict[int,Optional[Set[int]]] = dict((d, None) for d in difference)
 
         for pageno in intersection:
             my_page = self._pages[pageno]
             other_page = other._pages[pageno]
 
             if (my_page is None) ^ (other_page is None):
-                changes.update(pageno)
+                changes[pageno] = None
             elif my_page is None:
                 pass
             else:
-                if my_page.changed_bytes(other_page, page_addr=pageno * self.page_size):
-                    changes.add(pageno)
+                changed_offsets = my_page.changed_bytes(other_page, page_addr=pageno * self.page_size)
+                if changed_offsets:
+                    changes[pageno] = changed_offsets
 
         return changes
 
@@ -538,9 +568,65 @@ class PagedMemoryMixin(MemoryMixin):
         self._pages = new_page_dict
         return flushed
 
+
+class LabeledPagesMixin(PagedMemoryMixin):
+    def load_with_labels(self, addr: int, size: int=None, endness=None, **kwargs) -> Tuple[claripy.ast.Base,Tuple[Tuple[int,int,Any]]]:
+        if endness is None:
+            endness = self.endness
+
+        if type(size) is not int:
+            raise TypeError("Need size to be resolved to an int by this point")
+
+        if type(addr) is not int:
+            raise TypeError("Need addr to be resolved to an int by this point")
+
+        pageno, pageoff = self._divide_addr(addr)
+        vals = []
+
+        # fasttrack basic case
+        if pageoff + size <= self.page_size:
+            page = self._get_page(pageno, False, **kwargs)
+            vals.append(page.load(pageoff, size=size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))
+
+        else:
+            max_pageno = (1 << self.state.arch.bits) // self.page_size
+            bytes_done = 0
+            while bytes_done < size:
+                page = self._get_page(pageno, False, **kwargs)
+                sub_size = min(self.page_size-pageoff, size-bytes_done)
+                vals.append(page.load(pageoff, size=sub_size, endness=endness, page_addr=pageno*self.page_size, memory=self, cooperate=True, **kwargs))
+
+                bytes_done += sub_size
+                pageno = (pageno + 1) % max_pageno
+                pageoff = 0
+
+        labels = [ ]
+        out = self.PAGE_TYPE._compose_objects(vals, size, endness, memory=self, labels=labels, **kwargs)
+        l.debug("%s.load_with_labels(%#x, %d, %s) = %s", self.id, addr, size, endness, out)
+        return out, tuple(labels)
+
+
 class ListPagesMixin(PagedMemoryMixin):
     PAGE_TYPE = ListPage
 
 
+class MVListPagesMixin(PagedMemoryMixin):
+    PAGE_TYPE = MVListPage
+
+
+class ListPagesWithLabelsMixin(
+    LabeledPagesMixin,
+    ListPagesMixin,
+):
+    pass
+
+
+class MVListPagesWithLabelsMixin(
+    LabeledPagesMixin,
+    MVListPagesMixin,
+):
+    pass
+
+
 class UltraPagesMixin(PagedMemoryMixin):
     PAGE_TYPE = UltraPage
