@@ -1,7 +1,8 @@
+# pylint:disable=abstract-method
 import logging
 from typing import Optional, List, Set, Tuple
 
-import claripy
+from sortedcontainers import SortedSet
 
 from . import PageBase
 from angr.storage.memory_object import SimMemoryObject
@@ -12,20 +13,24 @@ l = logging.getLogger(name=__name__)
 
 
 class ListPage(MemoryObjectMixin, PageBase):
-    def __init__(self, memory=None, content=None, sinkhole=None, **kwargs):
+    def __init__(self, memory=None, content=None, sinkhole=None, mo_cmp=None, **kwargs):
         super().__init__(**kwargs)
 
         self.content: List[Optional[SimMemoryObject]] = content
+        self.stored_offset = SortedSet()
         if content is None:
             if memory is not None:
                 self.content: List[Optional[SimMemoryObject]] = [None] * memory.page_size  # TODO: this isn't the best
+        self._mo_cmp = mo_cmp
 
         self.sinkhole: Optional[SimMemoryObject] = sinkhole
 
     def copy(self, memo):
         o = super().copy(memo)
         o.content = list(self.content)
+        o.stored_offset = self.stored_offset.copy()
         o.sinkhole = self.sinkhole
+        o._mo_cmp = self._mo_cmp
         return o
 
     def load(self, addr, size=None, endness=None, page_addr=None, memory=None, cooperate=False, **kwargs):
@@ -63,8 +68,10 @@ class ListPage(MemoryObjectMixin, PageBase):
         size = global_end_addr - global_start_addr
         new_ast = self._default_value(global_start_addr, size, name='%s_%x' % (memory.id, global_start_addr), key=(self.category, global_start_addr), memory=memory, **kwargs)
         new_item = SimMemoryObject(new_ast, global_start_addr, endness=endness, byte_width=memory.state.arch.byte_width if memory is not None else 8)
-        for subaddr in range(global_start_addr - page_addr, addr):
+        subaddr_start = global_start_addr - page_addr
+        for subaddr in range(subaddr_start, addr):
             self.content[subaddr] = new_item
+            self.stored_offset.add(subaddr)
         result[-1] = (global_start_addr, new_item)
 
     def store(self, addr, data, size=None, endness=None, memory=None, cooperate=False, **kwargs):
@@ -77,13 +84,15 @@ class ListPage(MemoryObjectMixin, PageBase):
         else:
             for subaddr in range(addr, addr + size):
                 self.content[subaddr] = data
+                self.stored_offset.add(subaddr)
 
     def merge(self, others: List['ListPage'], merge_conditions, common_ancestor=None, page_addr: int=None,
-              memory=None):
+              memory=None, changed_offsets: Optional[Set[int]]=None):
 
-        changed_offsets = set()
-        for other in others:
-            changed_offsets |= self.changed_bytes(other, page_addr)
+        if changed_offsets is None:
+            changed_offsets = set()
+            for other in others:
+                changed_offsets |= self.changed_bytes(other, page_addr)
 
         all_pages: List['ListPage'] = [self] + others
         if merge_conditions is None:
@@ -114,13 +123,14 @@ class ListPage(MemoryObjectMixin, PageBase):
             mos = set(mo for mo, _ in memory_objects)
             mo_bases = set(mo.base for mo, _ in memory_objects)
             mo_lengths = set(mo.length for mo, _ in memory_objects)
+            endnesses = set(mo.endness for mo in mos)
 
             if not unconstrained_in and not (mos - merged_objects):
                 continue
 
             # first, optimize the case where we are dealing with the same-sized memory objects
-            if len(mo_bases) == 1 and len(mo_lengths) == 1 and not unconstrained_in:
-                our_mo = self.content[b]
+            if len(mo_bases) == 1 and len(mo_lengths) == 1 and not unconstrained_in and len(endnesses) == 1:
+                the_endness = next(iter(endnesses))
                 to_merge = [(mo.object, fv) for mo, fv in memory_objects]
 
                 # Update `merged_to`
@@ -139,7 +149,7 @@ class ListPage(MemoryObjectMixin, PageBase):
                 # new_object = self._replace_memory_object(our_mo, merged_val, page_addr, memory.page_size)
 
                 self.store(b,
-                           SimMemoryObject(merged_val, mo_base, endness='Iend_BE'),
+                           SimMemoryObject(merged_val, mo_base, endness=the_endness),
                            size=size,
                            cooperate=True
                            )
@@ -180,24 +190,34 @@ class ListPage(MemoryObjectMixin, PageBase):
                            )  # do not convert endianness again
                 merged_offsets.add(b)
 
+        self.stored_offset |= merged_offsets
         return merged_offsets
 
     def changed_bytes(self, other: 'ListPage', page_addr: int=None):
 
         candidates: Set[int] = set()
-        for i in range(len(self.content)):
-            if self._contains(i, page_addr):
-                candidates.add(i)
-        for i in range(len(other.content)):
-            if other._contains(i, page_addr):
-                candidates.add(i)
+        if self.sinkhole is None:
+            candidates |= self.stored_offset
+        else:
+            for i in range(len(self.content)):
+                if self._contains(i, page_addr):
+                    candidates.add(i)
+
+        if other.sinkhole is None:
+            candidates |= other.stored_offset
+        else:
+            for i in range(len(other.content)):
+                if other._contains(i, page_addr):
+                    candidates.add(i)
 
         byte_width = 8  # TODO: Introduce self.state if we want to use self.state.arch.byte_width
         differences: Set[int] = set()
         for c in candidates:
-            if not self._contains(c, page_addr) and other._contains(c, page_addr):
+            s_contains = self._contains(c, page_addr)
+            o_contains = other._contains(c, page_addr)
+            if not s_contains and o_contains:
                 differences.add(c)
-            elif self._contains(c, page_addr) and not other._contains(c, page_addr):
+            elif s_contains and not o_contains:
                 differences.add(c)
             else:
                 if self.content[c] is None:
@@ -206,14 +226,17 @@ class ListPage(MemoryObjectMixin, PageBase):
                 if other.content[c] is None:
                     other.content[c] = SimMemoryObject(other.sinkhole.bytes_at(page_addr+c, 1), page_addr + c,
                                                        byte_width=byte_width, endness='Iend_BE')
-                if self._contains(c, page_addr) and self.content[c] != other.content[c]:
-                    # Try to see if the bytes are equal
-                    self_byte = self.content[c].bytes_at(page_addr + c, 1)
-                    other_byte = other.content[c].bytes_at(page_addr + c, 1)
-                    if self_byte is not other_byte:
-                        #l.debug("%s: offset %x, two different bytes %s %s from %s %s", self.id, c,
-                        #        self_byte, other_byte,
-                        #        self[c].object.model, other[c].object.model)
+                if s_contains and self.content[c] != other.content[c]:
+                    same = None
+                    if self._mo_cmp is not None:
+                        same = self._mo_cmp(self.content[c], other.content[c], page_addr + c, 1)
+                    if same is None:
+                        # Try to see if the bytes are equal
+                        self_byte = self.content[c].bytes_at(page_addr + c, 1)
+                        other_byte = other.content[c].bytes_at(page_addr + c, 1)
+                        same = self_byte is other_byte
+
+                    if same is False:
                         differences.add(c)
                 else:
                     # this means the byte is in neither memory
