@@ -19,8 +19,8 @@ static const uint8_t PAGE_SHIFT = 12;
 
 typedef uint64_t address_t;
 typedef uint64_t unicorn_reg_id_t;
-typedef uint64_t vex_reg_offset_t;
-typedef uint64_t vex_tmp_id_t;
+typedef int64_t vex_reg_offset_t;
+typedef int64_t vex_tmp_id_t;
 
 enum taint_t: uint8_t {
 	TAINT_NONE = 0,
@@ -55,6 +55,15 @@ struct taint_entity_t {
 	std::vector<taint_entity_t> mem_ref_entity_list;
 	// Instruction in which the entity is used. Used for taint sinks; ignored for taint sources.
 	address_t instr_addr;
+	int64_t value_size;
+
+	taint_entity_t() {
+		reg_offset = -1;
+		tmp_id = -1;
+		mem_ref_entity_list.clear();
+		instr_addr = 0;
+		value_size = -1;
+	}
 
 	bool operator==(const taint_entity_t &other_entity) const {
 		if (entity_type != other_entity.entity_type) {
@@ -94,20 +103,24 @@ struct taint_entity_t {
 };
 
 // Hash function for unordered_map. Needs to be defined this way in C++.
-template <>
-struct std::hash<taint_entity_t> {
-	std::size_t operator()(const taint_entity_t &entity) const {
-		return entity.operator()(entity);
-	}
-};
+namespace std {
+	template <>
+	struct hash<taint_entity_t> {
+		std::size_t operator()(const taint_entity_t &entity) const {
+			return entity.operator()(entity);
+		}
+	};
+}
 
 struct memory_value_t {
 	uint64_t address;
     uint8_t value[MAX_MEM_ACCESS_SIZE];
     uint64_t size;
+	bool is_value_symbolic;
 
 	bool operator==(const memory_value_t &other_mem_value) const {
-		if ((address != other_mem_value.address) || (size != other_mem_value.size)) {
+		if ((address != other_mem_value.address) || (size != other_mem_value.size) ||
+			(is_value_symbolic != other_mem_value.is_value_symbolic)) {
 			return false;
 		}
 		return (memcmp(value, other_mem_value.value, size) == 0);
@@ -122,32 +135,63 @@ struct memory_value_t {
 
 struct mem_read_result_t {
 	std::vector<memory_value_t> memory_values;
-	bool is_value_symbolic;
+	bool is_mem_read_symbolic;
+	uint32_t read_size;
 };
 
 struct register_value_t {
 	uint64_t offset;
 	uint8_t value[MAX_REGISTER_BYTE_SIZE];
+	int64_t size;
+
+	bool operator==(const register_value_t &reg_value) const {
+		if (offset != reg_value.offset) {
+			return false;
+		}
+		return (memcmp(value, reg_value.value, MAX_REGISTER_BYTE_SIZE) == 0);
+	}
+
+	std::size_t operator()(const register_value_t &reg_value) const {
+		return std::hash<uint64_t>()(reg_value.offset);
+	}
+};
+
+template <>
+struct std::hash<register_value_t> {
+	std::size_t operator()(const register_value_t &value) const {
+		return value.operator()(value);
+	}
 };
 
 struct instr_details_t {
 	address_t instr_addr;
+	int64_t mem_write_addr;
+	int64_t mem_write_size;
 	bool has_concrete_memory_dep;
 	bool has_symbolic_memory_dep;
-	memory_value_t *memory_values;
-	uint64_t memory_values_count;
+	// Mark fields as mutable so that they can be updated after inserting into std::set
+	mutable memory_value_t *memory_values;
+	mutable uint64_t memory_values_count;
+	std::set<instr_details_t> instr_deps;
+	std::unordered_set<register_value_t> reg_deps;
+	std::vector<std::pair<address_t, uint64_t>> symbolic_mem_deps;
+
+	instr_details_t() {
+		has_concrete_memory_dep = false;
+		has_symbolic_memory_dep = false;
+		instr_deps.clear();
+		mem_write_addr = -1;
+		mem_write_size = -1;
+		reg_deps.clear();
+		symbolic_mem_deps.clear();
+	}
 
 	bool operator==(const instr_details_t &other_instr) const {
 		if ((instr_addr != other_instr.instr_addr) || (has_concrete_memory_dep != other_instr.has_concrete_memory_dep) ||
-			(has_symbolic_memory_dep != other_instr.has_symbolic_memory_dep) ||
-			(memory_values_count != other_instr.memory_values_count)) {
+			(has_symbolic_memory_dep != other_instr.has_symbolic_memory_dep) || (instr_deps != other_instr.instr_deps) ||
+			(reg_deps != other_instr.reg_deps)) {
 				return false;
 		}
-		for (uint64_t counter = 0; counter < memory_values_count; counter++) {
-			if (!(memory_values[counter] == other_instr.memory_values[counter])) {
-				return false;
-			}
-		}
 		return true;
 	}
 
@@ -160,13 +204,50 @@ struct block_details_t {
 	address_t block_addr;
 	uint64_t block_size;
 	std::vector<instr_details_t> symbolic_instrs;
-	std::vector<register_value_t> register_values;
 	bool vex_lift_failed;
 
 	void reset() {
 		block_addr = 0;
 		block_size = 0;
+		symbolic_instrs.clear();
 		vex_lift_failed = false;
+	}
+};
+
+// sym_block_details_t and sym_instr_details_t are used to store data, references to which are returned to state plugin
+struct sym_instr_details_t {
+	address_t instr_addr;
+	bool has_memory_dep;
+	memory_value_t *memory_values;
+	uint64_t memory_values_count;
+
+	bool operator==(const sym_instr_details_t &other_instr) const {
+		if ((instr_addr != other_instr.instr_addr) || (has_memory_dep != other_instr.has_memory_dep) ||
+			(memory_values_count != other_instr.memory_values_count)) {
+				return false;
+		}
+		for (auto counter = 0; counter < memory_values_count; counter++) {
+			if (!(memory_values[counter] == other_instr.memory_values[counter])) {
+				return false;
+			}
+		}
+		return true;
+	}
+
+	bool operator<(const sym_instr_details_t &other_instr) const {
+		return (instr_addr < other_instr.instr_addr);
+	}
+};
+
+struct sym_block_details_t {
+	address_t block_addr;
+	uint64_t block_size;
+	std::vector<sym_instr_details_t> symbolic_instrs;
+	std::vector<register_value_t> register_values;
+
+	void reset() {
+		block_addr = 0;
+		block_size = 0;
 		symbolic_instrs.clear();
 		register_values.clear();
 	}
@@ -174,10 +255,10 @@ struct block_details_t {
 
 // This struct is used only to return data to the state plugin since ctypes doesn't natively handle
 // C++ STL containers
-struct block_details_ret_t {
+struct sym_block_details_ret_t {
 	uint64_t block_addr;
     uint64_t block_size;
-    instr_details_t *symbolic_instrs;
+    sym_instr_details_t *symbolic_instrs;
     uint64_t symbolic_instrs_count;
     register_value_t *register_values;
     uint64_t register_values_count;
@@ -214,6 +295,8 @@ enum stop_t {
 	STOP_UNSUPPORTED_EXPR_UNKNOWN,
 	STOP_UNKNOWN_MEMORY_WRITE,
 	STOP_SYMBOLIC_MEM_DEP_NOT_LIVE,
+	STOP_SYSCALL_ARM,
+	STOP_SYMBOLIC_MEM_DEP_NOT_LIVE_CURR_BLOCK,
 };
 
 typedef std::vector<std::pair<taint_entity_t, std::unordered_set<taint_entity_t>>> taint_vector_t;
@@ -222,37 +305,40 @@ struct instruction_taint_entry_t {
 	// List of direct taint sources for a taint sink
 	taint_vector_t taint_sink_src_map;
 
-	// List of registers a taint sink depends on
-	std::unordered_set<taint_entity_t> dependencies_to_save;
+	// List of dependencies a taint sink depends on
+	std::unordered_map<taint_entity_enum_t, std::unordered_set<taint_entity_t>, std::hash<uint8_t>> dependencies;
 
 	// List of taint entities in ITE expression's condition, if any
 	std::unordered_set<taint_entity_t> ite_cond_entity_list;
 
-	// List of registers modified by instruction and whether register's final value depends on
+	// List of registers modified by instruction, their size and whether register's final value depends on
 	// it's previous value
-	std::vector<std::pair<vex_reg_offset_t, bool>> modified_regs;
+	std::vector<std::pair<vex_reg_offset_t, std::pair<int64_t, bool>>> modified_regs;
 
-	// Count of memory reads in the instruction
-	uint32_t mem_read_count;
+	// Count number of bytes read from memory by the instruction
+	uint32_t mem_read_size;
 
 	bool has_memory_read;
 	bool has_memory_write;
 
 	bool operator==(const instruction_taint_entry_t &other_instr_deps) const {
 		return (taint_sink_src_map == other_instr_deps.taint_sink_src_map) &&
-			   (dependencies_to_save == other_instr_deps.dependencies_to_save) &&
+			   (dependencies == other_instr_deps.dependencies) &&
 			   (has_memory_read == other_instr_deps.has_memory_read) &&
 			   (has_memory_write == other_instr_deps.has_memory_write);
 	}
 
 	void reset() {
-		dependencies_to_save.clear();
+		dependencies.clear();
+		dependencies.emplace(TAINT_ENTITY_MEM, std::unordered_set<taint_entity_t>());
+		dependencies.emplace(TAINT_ENTITY_REG, std::unordered_set<taint_entity_t>());
+		dependencies.emplace(TAINT_ENTITY_TMP, std::unordered_set<taint_entity_t>());
 		ite_cond_entity_list.clear();
 		taint_sink_src_map.clear();
 		modified_regs.clear();
 		has_memory_read = false;
 		has_memory_write = false;
-		mem_read_count = 0;
+		mem_read_size = 0;
 		return;
 	}
 };
@@ -260,6 +346,8 @@ struct instruction_taint_entry_t {
 struct block_taint_entry_t {
 	std::map<address_t, instruction_taint_entry_t> block_instrs_taint_data_map;
 	std::unordered_set<taint_entity_t> exit_stmt_guard_expr_deps;
+	// Track instruction that sets a VEX temp and list of VEX temps on which its value depends on
+	std::unordered_map<taint_entity_t, std::pair<address_t, std::unordered_set<taint_entity_t>>> vex_temp_deps;
 	address_t exit_stmt_instr_addr;
 	bool has_unsupported_stmt_or_expr_type;
 	stop_t unsupported_stmt_stop_reason;
@@ -267,6 +355,8 @@ struct block_taint_entry_t {
 
 	bool operator==(const block_taint_entry_t &other_entry) const {
 		return (block_instrs_taint_data_map == other_entry.block_instrs_taint_data_map) &&
+			   (vex_temp_deps == other_entry.vex_temp_deps) &&
+			   (exit_stmt_instr_addr == other_entry.exit_stmt_instr_addr) &&
 			   (exit_stmt_guard_expr_deps == other_entry.exit_stmt_guard_expr_deps) &&
 			   (block_next_entities == other_entry.block_next_entities);
 	}
@@ -279,23 +369,31 @@ struct stop_details_t {
 };
 
 struct processed_vex_expr_t {
-	std::unordered_set<taint_entity_t> taint_sources;
-	std::unordered_set<taint_entity_t> ite_cond_entities;
+	std::unordered_map<taint_entity_enum_t, std::unordered_set<taint_entity_t>, std::hash<uint8_t>> taint_sources;
+	std::unordered_map<taint_entity_enum_t, std::unordered_set<taint_entity_t>, std::hash<uint8_t>> ite_cond_entities;
 	bool has_unsupported_expr;
 	stop_t unsupported_expr_stop_reason;
-	uint32_t mem_read_count;
+	uint32_t mem_read_size;
+	int64_t value_size;
 
 	void reset() {
 		taint_sources.clear();
+		taint_sources.emplace(TAINT_ENTITY_MEM, std::unordered_set<taint_entity_t>());
+		taint_sources.emplace(TAINT_ENTITY_REG, std::unordered_set<taint_entity_t>());
+		taint_sources.emplace(TAINT_ENTITY_TMP, std::unordered_set<taint_entity_t>());
 		ite_cond_entities.clear();
+		ite_cond_entities.emplace(TAINT_ENTITY_MEM, std::unordered_set<taint_entity_t>());
+		ite_cond_entities.emplace(TAINT_ENTITY_REG, std::unordered_set<taint_entity_t>());
+		ite_cond_entities.emplace(TAINT_ENTITY_TMP, std::unordered_set<taint_entity_t>());
 		has_unsupported_expr = false;
-		mem_read_count = 0;
+		mem_read_size = 0;
+		value_size = -1;
 	}
 };
 
 struct instr_slice_details_t {
 	std::set<instr_details_t> dependent_instrs;
-	std::unordered_set<vex_reg_offset_t> concrete_registers;
+	std::unordered_map<vex_reg_offset_t, int64_t> concrete_registers;
 };
 
 struct CachedPage {
@@ -318,13 +416,12 @@ typedef std::unordered_set<vex_reg_offset_t> RegisterSet;
 typedef std::unordered_map<vex_reg_offset_t, unicorn_reg_id_t> RegisterMap;
 typedef std::unordered_set<vex_tmp_id_t> TempSet;
 
-struct mem_access_t {
+struct mem_write_t {
 	address_t address;
 	uint8_t value[MAX_MEM_ACCESS_SIZE]; // assume size of any memory write is no more than 8
 	int size;
-	int clean; // save current page bitmap
-	bool is_symbolic;
-}; // actually it should be `mem_write_t` :)
+	std::vector<taint_t> previous_taint;
+};
 
 struct mem_update_t {
 	address_t address;
@@ -353,7 +450,7 @@ class State {
 
 	uc_context *saved_regs;
 
-	std::vector<mem_access_t> mem_writes;
+	std::vector<mem_write_t> mem_writes;
 	// List of all memory writes and their taint status
 	// Memory write instruction address -> is_symbolic
 	// TODO: Need to modify memory write taint handling for architectures that perform multiple
@@ -368,10 +465,7 @@ class State {
 	std::unordered_map<address_t, instr_slice_details_t> instr_slice_details_map;
 	// List of instructions in a block that should be executed symbolically. These are stored
 	// separately for easy rollback in case of errors.
-	block_details_t block_details;
-
-	// List of registers which are concrete dependencies of a block's instructions executed symbolically
-	std::unordered_set<vex_reg_offset_t> block_concrete_dependencies;
+	block_details_t curr_block_details;
 
 	// List of register values at start of block
 	std::unordered_map<vex_reg_offset_t, register_value_t> block_start_reg_values;
@@ -382,13 +476,16 @@ class State {
 	RegisterSet block_symbolic_registers, block_concrete_registers;
 	TempSet block_symbolic_temps;
 
+	// List of instructions that should be executed symbolically
+	std::vector<block_details_t> blocks_with_symbolic_instrs;
+
 	// the latter part of the pair is a pointer to the page data if the page is direct-mapped, otherwise NULL
 	std::map<address_t, std::pair<taint_t *, uint8_t *>> active_pages;
 	//std::map<uint64_t, taint_t *> active_pages;
 	std::set<uint64_t> stop_points;
 
 	address_t taint_engine_next_instr_address, taint_engine_stop_mem_read_instruction;
-	uint32_t taint_engine_stop_mem_read_count;
+	uint32_t taint_engine_stop_mem_read_size;
 
 	address_t unicorn_next_instr_addr;
 	address_t prev_stack_top_addr;
@@ -404,13 +501,15 @@ class State {
 
 	std::pair<taint_t *, uint8_t *> page_lookup(address_t address) const;
 
-	std::unordered_set<taint_entity_t> compute_dependencies_to_save(const std::unordered_set<taint_entity_t> &taint_sources) const;
 	void compute_slice_of_instrs(address_t instr_addr, const instruction_taint_entry_t &instr_taint_entry);
+	void compute_slice_of_instrs_for_vex_temps(instr_details_t &instr);
 	instr_details_t compute_instr_details(address_t instr_addr, const instruction_taint_entry_t &instr_taint_entry);
 	void get_register_value(uint64_t vex_reg_offset, uint8_t *out_reg_value) const;
+	// Return list of all dependent instructions including dependencies of those dependent instructions
+	std::set<instr_details_t> get_list_of_dep_instrs(const instr_details_t &instr) const;
 
 	// Returns a pair (taint sources, list of taint entities in ITE condition expression)
-	processed_vex_expr_t process_vex_expr(IRExpr *expr, address_t instr_addr, bool is_exit_stmt);
+	processed_vex_expr_t process_vex_expr(IRExpr *expr, IRTypeEnv *vex_block_tyenv, address_t instr_addr, bool is_exit_stmt);
 
 	// Determine cumulative result of taint statuses of a set of taint entities
 	// EG: This is useful to determine the taint status of a taint sink given it's taint sources
@@ -420,14 +519,15 @@ class State {
 	// unordered_set
 	taint_status_result_t get_final_taint_status(const std::vector<taint_entity_t> &taint_sources) const;
 
+	int32_t get_vex_expr_result_size(IRExpr *expr, IRTypeEnv* tyenv) const;
+
 	bool is_block_exit_guard_symbolic() const;
 	bool is_block_next_target_symbolic() const;
-	bool is_symbolic_register(vex_reg_offset_t reg_offset) const;
+	bool is_symbolic_register(vex_reg_offset_t reg_offset, int64_t reg_size) const;
 	bool is_symbolic_temp(vex_tmp_id_t temp_id) const;
-	bool is_symbolic_register_or_temp(const taint_entity_t &entity) const;
 
-	void mark_register_symbolic(vex_reg_offset_t reg_offset, bool do_block_level);
-	void mark_register_concrete(vex_reg_offset_t reg_offset, bool do_block_level);
+	void mark_register_symbolic(vex_reg_offset_t reg_offset, int64_t reg_size);
+	void mark_register_concrete(vex_reg_offset_t reg_offset, int64_t reg_size);
 	void mark_temp_symbolic(vex_tmp_id_t temp_id);
 
 	block_taint_entry_t process_vex_block(IRSB *vex_block, address_t address);
@@ -435,6 +535,10 @@ class State {
 	void propagate_taints();
 	void propagate_taint_of_one_instr(address_t instr_addr, const instruction_taint_entry_t &instr_taint_entry);
 
+	// Save values of concrete memory reads performed by an instruction and it's dependencies
+	void save_concrete_memory_deps(instr_details_t &instr);
+	// Find address and size of all symbolic memory reads performed by an instruction and it's dependencies
+	std::vector<std::pair<address_t, uint64_t>> find_symbolic_mem_deps(const instr_details_t &instr) const;
 	void update_register_slice(address_t instr_addr, const instruction_taint_entry_t &curr_instr_taint_entry);
 
 	// Inline functions
@@ -447,22 +551,6 @@ class State {
 		return (blacklisted_registers.count(reg_offset) > 0);
 	}
 
-	inline vex_reg_offset_t get_full_register_offset(vex_reg_offset_t reg_offset) const {
-		auto vex_sub_reg_mapping_entry = vex_sub_reg_map.find(reg_offset);
-		if (vex_sub_reg_mapping_entry != vex_sub_reg_map.end()) {
-			return vex_sub_reg_mapping_entry->second;
-		}
-		if (reg_size_map.find(reg_offset) != reg_size_map.end()) {
-			return reg_offset;
-		}
-		for (auto &entry: reg_size_map) {
-			if ((entry.first <= reg_offset) && (reg_offset <= entry.first + entry.second)) {
-				return entry.first;
-			}
-		}
-		return reg_offset;
-	}
-
 	inline unsigned int arch_pc_reg_vex_offset() const {
 		switch (arch) {
 			case UC_ARCH_X86:
@@ -478,7 +566,7 @@ class State {
 		}
 	}
 
-	inline unsigned int arch_pc_reg() const {
+	inline int arch_pc_reg() const {
 		switch (arch) {
 			case UC_ARCH_X86:
 				return mode == UC_MODE_64 ? UC_X86_REG_RIP : UC_X86_REG_EIP;
@@ -493,7 +581,7 @@ class State {
 		}
 	}
 
-	inline unsigned int arch_sp_reg() const {
+	inline int arch_sp_reg() const {
 		switch (arch) {
 			case UC_ARCH_X86:
 				return mode == UC_MODE_64 ? UC_X86_REG_RSP : UC_X86_REG_ESP;
@@ -508,27 +596,12 @@ class State {
 		}
 	}
 
-	inline unsigned int arch_byte_width() const {
-		switch(arch) {
-			case UC_ARCH_X86:
-				return mode == UC_MODE_64 ? 8 : 4;
-			case UC_ARCH_ARM:
-				return 4;
-			case UC_ARCH_ARM64:
-				return 8;
-			case UC_ARCH_MIPS:
-				return mode == UC_MODE_MIPS64 ? 8 : 4;
-			default:
-				return -1;
-		}
-	}
-
 	inline bool is_thumb_mode() const {
 		// unicorn engine mode doesn't reflect the current execution mode correctly and so we check T bit in the CPSR
 		// register to determine if we are executing in ARM or THUMB mode.
 		uint32_t cpsr_reg_val;
 		uc_reg_read(uc, UC_ARM_REG_CPSR, &cpsr_reg_val);
-		return (cpsr_reg_val & 32 == 0);
+		return ((cpsr_reg_val & 32) != 0);
 	}
 
 	public:
@@ -558,8 +631,6 @@ class State {
 		RegisterSet symbolic_registers; // tracking of symbolic registers
 		RegisterSet blacklisted_registers;  // Registers which shouldn't be saved as a concrete dependency
 		RegisterMap vex_to_unicorn_map; // Mapping of VEX offsets to unicorn registers
-		RegisterMap vex_sub_reg_map; // Mapping of VEX sub-registers to their main register
-		std::unordered_map<vex_reg_offset_t, uint64_t> reg_size_map;
 		RegisterSet artificial_vex_registers; // Artificial VEX registers
 		std::unordered_map<vex_reg_offset_t, uint64_t> cpu_flags;	// VEX register offset and bitmask for CPU flags
 		int64_t cpu_flags_register;
@@ -568,8 +639,8 @@ class State {
 		// Result of all memory reads executed. Instruction address -> memory read result
 		std::unordered_map<address_t, mem_read_result_t> mem_reads_map;
 
-		// List of instructions that should be executed symbolically
-		std::vector<block_details_t> blocks_with_symbolic_instrs;
+		// List of instructions that should be executed symbolically; used to store data to return
+		std::vector<sym_block_details_t> block_details_to_return;
 
 		bool track_bbls;
 		bool track_stack;
@@ -602,7 +673,7 @@ class State {
 
 		uc_err start(address_t pc, uint64_t step = 1);
 
-		void stop(stop_t reason);
+		void stop(stop_t reason, bool do_commit=false);
 
 		void step(address_t current_address, int32_t size, bool check_stop_points=true);
 
@@ -674,22 +745,13 @@ class State {
 		}
 
 		inline bool is_symbolic_taint_propagation_disabled() const {
-			return (is_symbolic_tracking_disabled() || block_details.vex_lift_failed);
+			return (is_symbolic_tracking_disabled() || curr_block_details.vex_lift_failed);
 		}
 
 		inline address_t get_taint_engine_stop_mem_read_instr_addr() const {
 			return taint_engine_stop_mem_read_instruction;
 		}
 
-		inline uint32_t get_pending_mem_reads_count() const {
-			return taint_engine_stop_mem_read_count;
-		}
-
-		inline void decrement_pending_mem_reads_count() {
-			taint_engine_stop_mem_read_count--;
-			return;
-		}
-
 		inline void update_previous_stack_top() {
 			prev_stack_top_addr = get_stack_pointer();
 			return;
